/*
 * comled-1.0.0
 *--------------------
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 */


 /********************************************************************
    Compile with gcc -O2 comled.c -o comled


    RS232 pinout

      1   DCD       input
      2   RxD       input
      3   TxD       output
      4   DTR       output    used for connection LED
      5   GND
      6   DSR       input
      7   RTS       output    set to 1 to have power for switch(es)
      8   CTS       input
      9   RI        input


  Every second the presence of /var/ipc/red/active is tested and the LED
  (DTR pin 4) is set or reset accordingly.

  Switches can be connected to pins 1, 6, 8 and 9. Edge detection is done 
  every second and /etc/rc.d/rc.comled is launched with parameter to
  indicate which switch was switched
  

 ********************************************************************/


 /********************************************************************
  Programmer:  Michael Sturm
       michi@advancedart.de
       http://forestjump.no-ip.info
  
     Version:  comled-0.2
        Date:  01.06.2005

  modifications by weizen_42 email: cop at ban-solms dot de

  $Id: comled.c 163 2007-05-19 13:34:41Z weizen_42 $

  date          who
  ====================================================================
  2006-04-30    weizen_42     v0.9.0 lot's of stuff, most importantly get switches

  2007-02
  2007-03       weizen_42     v0.9.2/v0.9.3 changed blink frequency, also blink when dialing

 ********************************************************************/



#include <signal.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <syslog.h>
#include <time.h>
#include <unistd.h>
#include <asm/io.h>
#include <sys/stat.h>
#include <sys/types.h>


#define RED_ACTIVE          "/var/ipcop/red/active"
#define SCRIPT              "/etc/rc.d/rc.comled"

#define false              0
#define true               !false

#define D_COM1              0x3F8
#define D_COM2              0x2F8

#define D_MODEM_CTRL(_port)  (_port+4)
#define D_MODEM_SATUS(_port) (_port+6)

#define ENABLE_DTR        0x01        // 8250 DTR enable
#define ENABLE_RTS        0x02        // 8250 RTS enable
#define ENABLE_OUT1       0x04        // 8250 OUT1 enable
#define ENABLE_OUT2       0x08        // 8250 OUT2 enable
#define EDGE_CTS          0x01        // 8250 CTS edge detected
#define EDGE_DSR          0x02        // 8250 DSR edge detected
#define EDGE_RI           0x04        // 8250 RI falling edge
#define EDGE_DCD          0x08        // 8250 DCD edge detected
#define STATUS_CTS        0x10        // 8250 CTS is high
#define STATUS_DSR        0x20        // 8250 DSR is high
#define STATUS_RI         0x40        // 8250 RI is high
#define STATUS_DCD        0x80        // 8250 DCD is high

#define LED_ON            ENABLE_RTS | ENABLE_DTR
#define LED_OFF           ENABLE_RTS


enum STATES { STATE_DISCONNECTED = 0, STATE_DIALING, STATE_CONNECTED };

typedef void (*sighandler_t)(int);

static sighandler_t
handle_signal (int sig_nr, sighandler_t signalhandler) {
   struct sigaction neu_sig, alt_sig;

   neu_sig.sa_handler = signalhandler;
   sigemptyset (&neu_sig.sa_mask);
   neu_sig.sa_flags = SA_RESTART;
   if (sigaction (sig_nr, &neu_sig, &alt_sig) < 0)
      return SIG_ERR;
   return alt_sig.sa_handler;
}

static void start_daemon (void) {
   int i;
   pid_t pid;

   if (setsid() > 0) {
    printf("comled kann nicht Session-Fhrer werden!\n");
    exit (EXIT_FAILURE);
   }
   handle_signal (SIGHUP, SIG_IGN);
   if ((pid = fork ()) != 0)
    exit (EXIT_FAILURE);

   chdir ("/");
   umask (0);
   for (i = sysconf (_SC_OPEN_MAX); i > 0; i--)
    close (i);
}


int main( int argc, char *argv[] ) 
{
  int i;
  int iPort = -1;

  int bQuiet = false;
  int bDaemon = false;

  struct timespec sleep_blinkfast, sleep_blinkslow, sleep_result;
  int switches = 0;
  int counter = 0;
  int bToggle = 0;

  int connstate = STATE_DISCONNECTED;

  for(i = 1; i < argc; i++)
  {
    if ( !strcasecmp(argv[i], "COM1")  )
    {
      iPort = D_COM1;
    }
    if ( !strcasecmp(argv[i], "COM2")  )
    {
      iPort = D_COM2;
    }
    if ( !strcasecmp(argv[i], "-q") || !strcasecmp(argv[i], "--quiet") )
    {
      bQuiet = true;
    }
    if ( !strcasecmp(argv[i], "-d") || !strcasecmp(argv[i], "--daemon") )
    {
      bDaemon = true;
    }
  }

  if ( iPort == -1 )
  {
    printf( "\nusage: %s <COM1|COM2> [-d | --daemon] [-q | --quiet]\n", argv[0]);
    exit( 1 );
  }

  if ( ioperm(iPort, 8, 1) != 0 ) 
  {
    perror( "ERROR: can't open COM device for read/write" );
    return(-1);
  }

   
  if( bDaemon )  
    start_daemon();

  sleep_blinkfast.tv_sec = 0;
  sleep_blinkfast.tv_nsec = 50000000;

  sleep_blinkslow.tv_sec = 0;
  sleep_blinkslow.tv_nsec = 200000000;

  while(1) 
  {
    // play with LED to show we're active
    if ( counter >= 150 )
    {
      if( !bQuiet )
        printf("\nplay around");
      counter = 0;
      for (i = 0; i < 5 ; i++)
      {
        outb( LED_OFF, D_MODEM_CTRL(iPort) );   
        nanosleep(&sleep_blinkfast, &sleep_result);
        outb( LED_ON, D_MODEM_CTRL(iPort) );   
        nanosleep(&sleep_blinkfast, &sleep_result);
      }
    }

    switch ( connstate )
    {
      case STATE_DISCONNECTED:
        counter++;
        outb( LED_OFF, D_MODEM_CTRL(iPort) );   
        if( !bQuiet )
          printf("\noffline");
        break;

      case STATE_DIALING:
        counter = 0;
        if ( !bToggle )
        {
          outb( LED_OFF, D_MODEM_CTRL(iPort) );   
        }
        else
        {
          outb( LED_ON, D_MODEM_CTRL(iPort) );
        }
        bToggle = !bToggle;
        break;

      case STATE_CONNECTED:
        counter++;
        outb( LED_ON, D_MODEM_CTRL(iPort) );
        if( !bQuiet )
          printf("\nonline");
    }

    // determine state
    if ( access(RED_ACTIVE, F_OK) != -1 )
    {
      connstate = STATE_CONNECTED;
    }
    else if ( !system("/bin/ps -ef | /bin/grep -q '[p]ppd'") || !system("/bin/ps -ef | /bin/grep -q '[c]onnectioncheck'") )
    {
      connstate = STATE_DIALING;
    }
    else
    {
      connstate = STATE_DISCONNECTED;
    }

    // remove edges
    inb( D_MODEM_SATUS(iPort) );

    nanosleep(&sleep_blinkslow, &sleep_result);

    // get modem status and start scripts if edge detected
    switches = inb( D_MODEM_SATUS(iPort) );

    if ( switches & EDGE_CTS )
    {
      system(SCRIPT " CTS");
    }
    if ( switches & EDGE_DSR )
    {
      system(SCRIPT " DSR");
    }
    if ( switches & EDGE_DCD )
    {
      system(SCRIPT " DCD");
    }
    if ( switches & (EDGE_RI | STATUS_RI) )
    {
      system(SCRIPT " RI");
    }
  }
  return EXIT_SUCCESS;
}
